/* SPDX-License-Identifier: MIT */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <unistd.h>

#ifndef RENAME_EXCHANGE
# define RENAME_EXCHANGE   (1 << 1)
#endif

void show_help(int ret);
char* xvasprintf(const char* fmt, va_list) __attribute__((format(printf, 1, 0))) __attribute__((malloc));
char* xasprintf(const char* fmt, ...) __attribute__((format(printf, 1, 2))) __attribute__((malloc));
void log_verbose(const char* fmt, ...) __attribute__((format(printf, 1, 2)));
void log_error(const char* fmt, ...) __attribute__((format(printf, 1, 2)));
void log_error_errno(const char* descr, int errnum);

int rename_exchange(const char* source, const char* dest);
int fallback_3way_exchange(const char* source, const char* dest);

static const char* arg0 = NULL;
static struct option long_options[] = {
	{"verbose", no_argument,       0,  'v' },
	{"help",    no_argument,       0,  255 },
	{0,         0,                 0,  0 }
};
static int verbose = 0;

int main(int argc, char* argv[])
{
	int rc;

	arg0 = argv[0];
	while ((rc = getopt_long(argc, argv, "v", long_options, NULL)) != -1) {
		switch (rc) {
		case 'v':
			++verbose;
			break;
		case 255:
			show_help(EXIT_SUCCESS);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc < 2) {
		/* Too few arguments. */
		show_help(EXIT_FAILURE);
	}

	if (argc > 2) {
		/* Too many arguments. */
		show_help(EXIT_FAILURE);
	}

	const char* source = argv[0];
	const char* target = argv[1];
	rc = rename_exchange(source, target);
	if (rc >= 0) {
		exit(EXIT_SUCCESS);
	}
	if (rc < 0) {
		/* ENOSYS is returned if there is no renameat2 syscall.
		 * Also, according to renameat2(2) manual page:
		 * - EINVAL is returned in case underlying FS does not
		 *   support RENAME_EXCHANGE.
		 * In all these cases we have to attempt a classic
		 * 3-way name exchange.
		 */
		if (
		errno != EINVAL &&
		errno != ENOSYS) {
			log_error_errno("renameat2", errno);
			exit(EXIT_FAILURE);
		}
	}

	rc = fallback_3way_exchange(source, target);
	if (rc < 0) {
		log_error_errno("3-way rename", errno);
		exit(EXIT_FAILURE);
	}

	return 0;
}

void show_help(int ret)
{
	printf("Usage: %s SOURCE DEST\n"
	       "", arg0);
	puts("Exchange names of SOURCE and DEST.");
	exit(ret);
}

/* Die in case of error condition on vasprintf(3).
 * Returns the newly allocated string.
 */
char* xvasprintf(const char* fmt, va_list ap)
{
	char* r;

	if (vasprintf(&r, fmt, ap) < 0) {
		fprintf(stderr, "%s: %s\n", arg0, "Out of memory");
		exit(EXIT_FAILURE);
	}
	return r;
}

char* xasprintf(const char* fmt, ...)
{
	char* r;

	va_list ap;
	va_start(ap, fmt);

	r = xvasprintf(fmt, ap);

	va_end(ap);
	return r;
}

void log_error(const char* fmt, ...)
{
	char* msg;

	fflush(stdout);

	va_list ap;
	va_start(ap, fmt);
	msg = xvasprintf(fmt, ap);
	fprintf(stderr, "%s: %s\n", arg0, msg);
	free(msg);
	va_end(ap);
}

/* To be used instead of perror(3). */
void log_error_errno(const char* descr, int errnum)
{
	log_error("%s: %s (%s)",
		descr, strerrorname_np(errnum), strerror(errnum));
}

void log_verbose(const char* fmt, ...)
{
	if (verbose <= 0)
		return;

	char* msg;

	va_list ap;
	va_start(ap, fmt);
	msg = xvasprintf(fmt, ap);
	fprintf(stdout, "%s: %s\n", arg0, msg);
	free(msg);
	va_end(ap);
}

int rename_exchange(const char* source, const char* dest) {
	int rc, saved_errno;
	rc = syscall(SYS_renameat2,
			AT_FDCWD, source, AT_FDCWD, dest,
			RENAME_EXCHANGE);
	if (rc >= 0) {
		saved_errno = errno;
		/* Mimic mv's verbose reporting. */
		log_verbose("exchanged %s <-> %s", source, dest);
		errno = saved_errno;
	}
	return rc < 0 ? -1 : 0;
}

/* Rename @source to a new, securely obtained temporary filename defined
 * by the pattern in @tmpname.
 *
 * Create the correct object behind a temporary filename
 * for rename(2) to work: directory if @sibling is a directory,
 * otherwise a regular file.
 *
 * We do not really need an open descriptor for the new dir, since mkdir(2)
 * implies the effect of O_EXCL, but we do need one with O_EXCL for the new
 * file. Take care in both cases.
 */
static int rename_away(const char* source, char* tmpname) {
	int rc;
	int saved_errno;
	int fd = -1;
	struct stat st_source;

	rc = lstat(source, &st_source);
	if (rc < 0) {
		log_error_errno(source, errno);
		return rc;
	}
	if ((st_source.st_mode & S_IFMT) == S_IFDIR) {
		rc = mkdtemp(tmpname) != NULL ? 0 : -1;
	} else {
		rc = fd = mkstemp(tmpname);
	}
	if (rc < 0) {
		log_error_errno("mkstemp", errno);
		return rc;
	}

	rc = renameat(AT_FDCWD, source, AT_FDCWD, tmpname);
	saved_errno = errno;
	if (fd >= 0)
		close(fd);
	errno = saved_errno;
	return rc;
}

/* Exchange two pathnames, @source and @dest, without resorting
 * to renameat2(2).
 */
int fallback_3way_exchange(const char* source, const char* dest) {
	int rc;
	char* tmpname;
	tmpname = xasprintf("%s.xchg~%d~.XXXXXX", dest, getpid());

	rc = rename_away(dest, tmpname);
	if (rc < 0) {
		log_error_errno("renameat", errno);
		rc = -1;
		goto out;
	}
	/* Mimic mv's verbose reporting. */
	log_verbose("renamed %s -> %s", dest, tmpname);

	rc = renameat(AT_FDCWD, source, AT_FDCWD, dest);
	if (rc < 0) {
		log_error_errno("renameat", errno);
		rc = -1;
		goto out;
	}
	/* Mimic mv's verbose reporting. */
	log_verbose("renamed %s -> %s", source, dest);

	rc = renameat(AT_FDCWD, tmpname, AT_FDCWD, source);
	if (rc < 0) {
		log_error_errno("renameat", errno);
		rc = -1;
		goto out;
	}
	/* Mimic mv's verbose reporting. */
	log_verbose("renamed %s -> %s", tmpname, source);

	rc = 0;
out:
	free(tmpname);
	return rc;
}
