/*
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Adapted from FreeBSD implementation of realpath(3), which has
 * copyright (c) 2003 Constantin S. Svintsoff <kostik@iclub.nsu.ru>
 *
 * This program "almost canonicalizes" a file path.
 * The algorithm is similar to realpath(3), with the following amendments:
 * - allow any component to be missing;
 * - the last path component is ignored if it is reached as a result of
 *   following a symlink in the last component of the original path;
 *
 * If it ever gains a sophisticated CLI with options, or its functionality is
 * merged into an existing general-purpose tool, it will be installed under a
 * different name.
 */

#include <sys/param.h>
#include <sys/stat.h>

#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
 * Find the real name of path, by removing all ".", ".." and symlink
 * components.  Returns @resolved on success, or NULL on failure,
 * in which case the path which caused trouble is left in resolved.
 */
static char *
realpath_once1(const char * path, char * resolved)
{
	struct stat sb;
	char *p, *q;
	size_t left_len, next_token_len, resolved_len;
	unsigned symlinks;
	unsigned tail_symlinks = 0;
	ssize_t slen;
	char left[PATH_MAX], next_token[PATH_MAX], symlink[PATH_MAX];

	symlinks = 0;
	if (path[0] == '/') {
		resolved[0] = '/';
		resolved[1] = '\0';
		if (path[1] == '\0')
			return resolved;
		resolved_len = 1;
		left_len = strlcpy(left, path + 1, sizeof(left));
	} else {
		if (getcwd(resolved, PATH_MAX) == NULL) {
			resolved[0] = '.';
			resolved[1] = '\0';
			return NULL;
		}
		resolved_len = strlen(resolved);
		left_len = strlcpy(left, path, sizeof(left));
	}
	if (left_len >= sizeof(left) || resolved_len >= PATH_MAX) {
		errno = ENAMETOOLONG;
		return NULL;
	}

	/*
	 * Iterate over path components in `left'.
	 */
	while (left_len != 0) {
		/*
		 * Extract the next path component and adjust `left'
		 * and its length.
		 */
		p = strchr(left, '/');

		next_token_len = p != NULL ? (size_t)(p - left) : left_len;
		memcpy(next_token, left, next_token_len);
		next_token[next_token_len] = '\0';

		if (p != NULL) {
			left_len -= next_token_len + 1;
			memmove(left, p + 1, left_len + 1);
		} else {
			left[0] = '\0';
			left_len = 0;
		}

		if (resolved[resolved_len - 1] != '/') {
			if (resolved_len + 1 >= PATH_MAX) {
				errno = ENAMETOOLONG;
				return NULL;
			}
			resolved[resolved_len++] = '/';
			resolved[resolved_len] = '\0';
		}
		if (next_token[0] == '\0') {
			/* Handle consequential slashes. */
			continue;
		} else if (strcmp(next_token, ".") == 0) {
			continue;
		} else if (strcmp(next_token, "..") == 0) {
			/*
			 * Strip the last path component except when we have
			 * single "/"
			 */
			if (resolved_len > 1) {
				resolved[resolved_len - 1] = '\0';
				q = strrchr(resolved, '/') + 1;
				*q = '\0';
				resolved_len = q - resolved;
			}
			continue;
		}

		/*
		 * Append the next path component and lstat() it.
		 */
		resolved_len = strlcat(resolved, next_token, PATH_MAX);
		if (resolved_len >= PATH_MAX) {
			errno = ENAMETOOLONG;
			return NULL;
		}
		if (p == NULL && tail_symlinks > 0)
			/* Difference from realpath(3):
			 * like `readlink -m`, ignore last component, but only if a
			 * tailing symlink was once followed.
			 */
			break;
		if (lstat(resolved, &sb) != 0) {
			/* Difference from realpath(3):
			 * allow any component to be missing.
			 */
			if (errno == ENOENT)
				continue;
			return NULL;
		}
		if (S_ISLNK(sb.st_mode)) {
			if (symlinks++ > MAXSYMLINKS) {
				errno = ELOOP;
				return NULL;
			}
			slen = readlink(resolved, symlink, sizeof(symlink));
			if (slen <= 0 || slen >= (ssize_t)sizeof(symlink)) {
				if (slen < 0)
					; /* keep errno from readlink(2) call */
				else if (slen == 0)
					errno = ENOENT;
				else
					errno = ENAMETOOLONG;
				return NULL;
			}
			symlink[slen] = '\0';
			if (symlink[0] == '/') {
				resolved[1] = '\0';
				resolved_len = 1;
			} else {
				/* Strip the last path component. */
				q = strrchr(resolved, '/') + 1;
				*q = '\0';
				resolved_len = q - resolved;
			}

			/*
			 * If there are any path components left, then
			 * append them to symlink. The result is placed
			 * in `left'.
			 */
			if (p != NULL) {
				if (symlink[slen - 1] != '/') {
					if (slen + 1 >= (ssize_t)sizeof(symlink)) {
						errno = ENAMETOOLONG;
						return NULL;
					}
					symlink[slen] = '/';
					symlink[slen + 1] = 0;
				}
				left_len = strlcat(symlink, left,
				    sizeof(symlink));
				if (left_len >= sizeof(symlink)) {
					errno = ENAMETOOLONG;
					return NULL;
				}
			} else
				/* Difference from realpath(3):
				 * Record if the symlink to be followed is in the last
				 * component.
				 */
				tail_symlinks++;
			left_len = strlcpy(left, symlink, sizeof(left));
		} else if (!S_ISDIR(sb.st_mode) && p != NULL) {
			errno = ENOTDIR;
			return NULL;
		}
	}

	/*
	 * Remove trailing slash except when the resolved pathname
	 * is a single "/".
	 */
	if (resolved_len > 1 && resolved[resolved_len - 1] == '/')
		resolved[resolved_len - 1] = '\0';
	return resolved;
}

static char *
realpath_once(const char * restrict path, char * restrict resolved)
{
	char *m, *res;

	if (path == NULL) {
		errno = EINVAL;
		return NULL;
	}
	if (path[0] == '\0') {
		errno = ENOENT;
		return NULL;
	}
	if (resolved != NULL) {
		m = NULL;
	} else {
		m = resolved = malloc(PATH_MAX);
		if (resolved == NULL)
			return NULL;
	}
	res = realpath_once1(path, resolved);
	if (res == NULL)
		free(m);
	return res;
}

int
main(int argc, char * argv[])
{
	if (argc != 2)
		error(EXIT_FAILURE, 0, "Exactly 1 argument expected. Usage: $0 <path>");

	const char * source = argv[1];

	char* res = realpath_once(source, NULL);
	if (!res)
		error(EXIT_FAILURE, errno, "failed to resolve \"%s\"", source);

	printf("%s\n", res);
	free(res);
	return 0;
}
